/*
 * Copy me if you can.
 * by 20h
 */

#include <u.h>
#include <libc.h>
#include <draw.h>
#include <event.h>

#undef PIO2
#define PIO2 3.14159265358979323846e0 / 2

typedef struct Geopos Geopos;
struct Geopos {
	double lat, lon;
};

typedef struct Pos Pos;
struct Pos {
	Geopos cur, ocur;
	double dir, alt;
	char *str;
};

typedef struct Map Map;
struct Map {
	Image *bg;
	double alt;
};

typedef struct Posl Posl;
struct Posl {
	Geopos g;
	Posl *n, *p;
};

Image *backc;
Image *pbord;
Image *pinne;
Image *textc;
Image *waypc;

int alts[] = {1, 3, 6, 12, 25, 50, 150, 800, 2000, 7000, 12000};

Pos *pos, *view;
Map *bg;
Posl *posl;
Image *backbuf;
int nalt, nposl;
char flags;

enum {
	PFLAG = 0x1,
	NFLAG = 0x2,
	DFLAG = 0x4,
	WFLAG = 0x8
};

char *buttons[] =
{
	"zoom +",
	"zoom -",
	"refresh",
	"pos",
	"waypoint",
	"clr wayp",
	"exit",
	0,
};

enum {
	ZOOMIN,
	ZOOMOUT,
	REFRESH,
	POS,
	WAYPOINT,
	CLR_WAYPOINT,
	EXIT,
};

Menu menu =
{
	buttons
};

double
rad(double i)
{

	return (i * PI) / 180;
}

double
arad(double i)
{

	return (i * 180) / PI;
}

void
free_posl(Posl *p)
{

	if(p != nil)
	{
		while(p->n != nil)
		{
			p = p->n;
			free(p->p);
		}
		free(p);
	}
	return;
}

void
free_pos(Pos *p)
{

	if(p != nil)
	{
		if(p->str != nil)
			free(p->str);
		free(p);
	}
	return;
}

void
free_map(Map *m)
{

	if(m != nil)
	{
		if(m->bg != nil)
			freeimage(m->bg);
		free(m);
	}
	return;
}

void *
reallocg(void *v, ulong n)
{

	v = realloc(v, n);
	if(v == nil)
		sysfatal("no memory allocable.");

	return v;
}

char *
strdupg(char *s)
{
	char *ret;

	ret = reallocg(nil, strlen(s) + 1);
	strcpy(ret, s);
	ret[strlen(s)] = '\0';

	return ret;
}

Posl *
add_wayp(Posl *p, Geopos g)
{
	Posl *po, *ret;

	ret = p;

	po = reallocg(nil, sizeof(Posl));
	po->n = nil;
	po->p = nil;
	po->g = g;

	nposl++;
	if(p == nil)
	{
		ret = po;
	} else {
		if(nposl >= 100)
		{
			ret = p->n;
			ret->p = nil;
			p->n = nil;
			free_posl(p);
			nposl = 99;
			p = ret;
		}
		while(p->n != nil)
			p = p->n;
		p->n = po;
		po->p = p;
	}

	return ret;
}

Point
get_pos(Pos *po, Map *m, Geopos ge)
{
	Point ret;
	double km, wp[3], rot[3][3], re[3], rlon, rlat;
	int i;

	rlat = rad(po->cur.lat);
	rlon = rad(-po->cur.lon);

	re[0] = rad(ge.lat);
	re[1] = rad(90.0 - ge.lon);
	wp[0] = cos(re[0]) * cos(re[1]);
	wp[1] = sin(re[0]);
	wp[2] = cos(re[0]) * sin(re[1]);
	re[0] = re[1] = re[2] = 0;

	rot[0][0] = cos(rlon);
	rot[0][1] = 0.0;
	rot[0][2] = sin(rlon);
	rot[1][0] = sin(rlat) * sin(rlon);
	rot[1][1] = cos(rlat);
	rot[1][2] = -sin(rlat) * cos(rlon);
	rot[2][0] = cos(rlat) * -sin(rlon);
	rot[2][1] = sin(rlat);
	rot[2][2] = cos(rlat) * cos(rlon);

	i = -1;
	while(++i < 3)
	{
		re[0] += wp[i] * rot[0][i];
		re[1] += wp[i] * rot[1][i];
		re[2] += wp[i] * rot[2][i];
	}

	km = (po->alt * 0.86) / (6378.2 * (Dx(m->bg->r) >= Dy(m->bg->r) ?
					Dy(m->bg->r) : Dx(m->bg->r)));

	ret.x = re[0] / km + Dx(m->bg->r) / 2;
	ret.y = -re[1] / km + Dy(m->bg->r) / 2;

	return ret;
}

Geopos
get_geo(Pos *po, Map *m, double xlon, double ylat)
{
	Geopos ret;
	double km, wp[3], rot[3][3], re[3], rlat, rlon;
	int i;

	re[0] = re[1] = re[2] = 0;

	km = (po->alt * 0.86) / (6378.2 * (Dx(m->bg->r) >= Dy(m->bg->r) ?
					Dy(m->bg->r) : Dx(m->bg->r)));

	wp[0] = xlon * km;
	wp[1] = ylat * km;

	rlat = 1.0 - (wp[0] * wp[0] + wp[1] * wp[1]);
	if(rlat < 0)
	{
		ret.lat = 1024;
		ret.lon = 1024;
		return ret;
	} 
	wp[2] = sqrt(rlat);

	rlat = rad(-po->cur.lat);
	rlon = rad(po->cur.lon);

	rot[0][0] = cos(rlon);
	rot[0][1] = sin(rlat) * sin(rlon);
	rot[0][2] = cos(rlat) * sin(rlon);
	rot[1][0] = 0.0;
	rot[1][1] = cos(rlat);
	rot[1][2] = -sin(rlat);
	rot[2][0] = -sin(rlon);
	rot[2][1] = sin(rlat) * cos(rlon);
	rot[2][2] = cos(rlat) * cos(rlon);

	i = -1;
	while(++i < 3)
	{
		re[0] += wp[i] * rot[0][i];
		re[1] += wp[i] * rot[1][i];
		re[2] += wp[i] * rot[2][i];
	}

	ret.lat = arad(asin(re[1]));
	ret.lon = 90.0 - arad(atan2(re[2], re[0]));

	if(ret.lon > 180)
		ret.lon -= 360;

	return ret;
}

Geopos
get_gps(char *path)
{
	Geopos ret;
	int fd;
	char buf[128], *c, *d;

	snprint(buf, 127, "%s/position", path);
	fd = open(buf, OREAD);
	if(fd < 0)
		goto gps_is_bad;

	read(fd, buf, 127);
	close(fd);

	c = strchr(buf, 'Z');
	if(c == nil)
		goto gps_is_bad;
	c += 2;
	d = strchr(++c, '\t');
	if(d == nil)
		goto gps_is_bad;
	ret.lon = atof(d);
	c = strchr(d, '.');
	if(c == nil)
		goto gps_is_bad;
	d = strchr(c, '\t');
	if(d == nil)
		goto gps_is_bad;
	ret.lat = atof(d);

	return ret;
gps_is_bad:
	ret.lat = 1024;
	ret.lon = 1024;

	return ret;
}

Map *
get_map(Pos *po, Rectangle r)
{
	Image *m;
	Map *ret, *ma;
	int p[2];
	char buf[512], *lang;
	Point poin;

	if(pipe(p) < 0)
		return nil;

	lang = "EUR0407";
	if(po->cur.lon <= -35)
		lang = "USA0409";

	snprint(buf, sizeof(buf), "hget 'http://www.expedia.de/pub/agent.dll?qscr=mrdt&ID=3kQaz.&CenP=%f,%f&Lang=%s&Alti=%f&Size=%d,%d&Offs=0.000000,0.000000' | gif -9", po->cur.lat, po->cur.lon, lang, po->alt, Dx(r) + 100, Dy(r) + 100);

	if(fork() == 0)
	{
		close(p[0]);
		dup(p[1], 1);
		execl("/bin/rc", "rc", "-c", buf, 0);
		exits("execl");
	}

	m = readimage(display, p[0], 0);
	wait();
	close(p[0]);
	close(p[1]);

	ma = reallocg(nil, sizeof(Map));
	ma->bg = m;
	ma->alt = po->alt;

	ret = reallocg(nil, sizeof(Map));
	ret->bg = allocimage(display, r, RGB24, 0, DNofill);
	ret->alt = po->alt;

	poin = subpt(addpt(ma->bg->r.min, Pt(Dx(ma->bg->r) / 2, Dy(ma->bg->r) / 2)), Pt(Dx(ret->bg->r) / 2, Dy(ret->bg->r) / 2));
//	poin = get_pos(po, ma, get_geo(po, ma, -Dx(ret->bg->r) / 2, Dy(ret->bg->r) / 2));
	drawop(ret->bg, ret->bg->r, ma->bg, nil, poin, S);

	free_map(ma);

	return ret;
}

int
is_inside(Map *m, Geopos g)
{
	Point po;

	po = get_pos(view, m, g);
	if(po.x >= m->bg->r.min.x || po.x <= m->bg->r.max.x ||
		po.y >= m->bg->r.min.y || po.y <= m->bg->r.max.y)
			return 1;
	return 0;
}

void
draw_wayp(Posl *p, Map *m)
{
	Point po0, po1;

	if(p == nil)
		return;

	while(p != nil)
	{
		if(is_inside(m, p->g))
		{
			if(p->p != nil)
			{
				po0 = get_pos(view, m, p->p->g);
				po1 = get_pos(view, m, p->g);

				line(backbuf, po0, po1, 0, 0, 0, waypc, ZP);
			}
			if(p->n != nil && !is_inside(m, p->n->g))
			{
				po0 = get_pos(view, m, p->g);
				po0 = get_pos(view, m, p->n->g);

				line(backbuf, po0, po1, 0, 0, 0, waypc, ZP);
			}
		}
		p = p->n;
	}
	return;
}
			
void
draw_pos(Pos *p, Map *m)
{
	Point po, pol[4];
	double dx, dy;
	int radi;
	char str[128];

	radi = 15;

	po = get_pos(view, m, p->cur);
	if(p->dir >= 0)
	{
		pol[3].x = pol[0].x = po.x + radi * (2 / 3 + 1) * cos(rad(p->dir));
		pol[3].y = pol[0].y = po.y - radi * (2 / 3 + 1) * sin(rad(p->dir));

		dy = radi * 1 / 3 * (pol[0].x - po.x) / sqrt((pol[0].x - po.x) *
				(pol[0].x - po.x) + (pol[0].y - po.y) * (pol[0].y - po.y));
		dx = radi * 1 / 3 * (pol[0].y - po.y) / sqrt((pol[0].x - po.x) *
				(pol[0].x - po.x) + (pol[0].y - po.y) * (pol[0].y - po.y));
		pol[1].x = po.x + dx;
		pol[1].y = po.y - dy;
		pol[2].x = po.x - dx;
		pol[2].y = po.y + dy;
	}

	ellipse(backbuf, po, radi * 1 / 3, radi * 1 / 3, 1, pbord, ZP);
	if(p->dir >= 0)
		fillpoly(backbuf, pol, 4, 1, pinne, ZP);
	ellipse(backbuf, po, radi, radi, 0, pbord, ZP);
	if(p->dir >= 0)
		poly(backbuf, pol, 4, 0, 0, 0, pbord, pol[0]);
	fillellipse(backbuf, po, radi * 1 / 3 - 1, radi * 1 / 3 - 1, pinne, ZP);

	if(flags & PFLAG)
	{
		snprint(str, 128, "%f %f", p->cur.lon, p->cur.lat);
		string(backbuf, Pt(po.x - stringwidth(font, str) / 2,
				po.y + 15), textc, ZP, font, str);
	}

	return;
}

void
draw_nav(int val)
{
	double dlat, dlon, clat, clon;
	Geopos cac;
	Point po;

	if((val != 1 || bg == nil) && !(flags & NFLAG))
	{
		cac = get_gps("/mnt/gps");
		if(cac.lon <= 180 && cac.lon >= -180 && cac.lat <= 90 && cac.lat >= -90)
		{
			pos->cur = cac;
			if(bg == nil)
				view->cur = pos->cur;
		}
	}
	view->alt = alts[nalt];

	if(pos->ocur.lat != pos->cur.lat || pos->ocur.lon != pos->cur.lon
		|| bg == nil || view->alt != pos->alt || val == 1)
	{
		if(bg != nil) {
			po = get_pos(view, bg, pos->cur);
			dlat = Dx(bg->bg->r);
			dlon = Dy(bg->bg->r);
			clat = clon = 0;

			if(po.x < bg->bg->r.min.x + dlon / 8)
				clon = -dlon / 3;
			if(po.x > bg->bg->r.max.x - dlon / 8)
				clon = dlon / 3;
			if(po.y < bg->bg->r.min.y + dlat / 8)
				clat = dlat / 3;
			if(po.y > bg->bg->r.max.y - dlat / 8)
				clat = -dlat / 3;

			if(po.x < bg->bg->r.min.x || po.x > bg->bg->r.max.x ||
				po.y < bg->bg->r.min.y || po.y > bg->bg->r.max.y) {
					view->cur = pos->cur;
			} else {
				if(clon != 0 || clat != 0)
					view->cur = get_geo(view, bg, clon, clat);
			}
		}
	
		if(bg == nil || view->cur.lat != view->ocur.lat ||
			view->cur.lon != view->ocur.lon || view->alt != pos->alt ||
			val == 1)
		{
			if(bg != nil)
				free_map(bg);
			bg = get_map(view, rectsubpt(backbuf->r, backbuf->r.min));

			pos->alt = view->alt;
			view->ocur = view->cur;
		}

		dlat = pos->cur.lat - pos->ocur.lat;
		dlon = pos->cur.lon - pos->ocur.lon;

		if(pos->cur.lat == pos->ocur.lat)
		{
			if(pos->cur.lon - pos->ocur.lon < 0)
			{
				pos->dir = 180;
			} else {
				pos->dir = 0;
			}
		} else {
			if(pos->cur.lon == pos->ocur.lon)
			{
				if(pos->cur.lat - pos->ocur.lat < 0)
				{
					pos->dir = 270;
				} else {
					pos->dir = 90;
				}
			} else {
				pos->dir = arad(atan((pos->cur.lon - pos->ocur.lon) /
								(pos->cur.lat - pos->ocur.lat)));
				if(dlon < 0)
					pos->dir += 180;
				if(dlon > 0 && dlat < 0)
					pos->dir += 360;
			}
		}

		if(flags & WFLAG && (pos->cur.lat != pos->cur.lat || pos->cur.lon != pos->ocur.lon || posl == nil))
			posl = add_wayp(posl, pos->cur);

		pos->ocur = pos->cur;
	} else {
		pos->dir = -1;
	}

	drawop(backbuf, backbuf->r, bg->bg, nil, bg->bg->r.min, S);
	draw_pos(pos, bg);
	draw_wayp(posl, bg);

	draw(screen, screen->r, backbuf, nil, ZP);
	flushimage(display, 1);

	return;
}

void
eresized(int new)
{

	if(new && getwindow(display, Refnone) < 0)
		sysfatal("can't reattach to window");

	if(backbuf != nil)
		freeimage(backbuf);
	backbuf = allocimage(display, rectsubpt(screen->r, screen->r.min), screen->chan, 0, DWhite);
	if(backbuf == nil)
		exits("allocimage failed");

	draw_nav(1);

	return;
}

void
usage(void)
{

	print("usage: gmap [-pdw] [-n lon?lat] [-s sec] [-a alt]\n");
	exits(0);
}

void
main(int argc, char *argv[])
{
	int sec, Etimer;
	Event e;
	double lat, lon;
	char *test, *b;

	sec = 2;
	nalt = 5;
	lon = 13.377159;
	lat = 52.515986;

	ARGBEGIN {
		case 'd':
			flags |= PFLAG | DFLAG;
			break;
		case 's':
			sec = atoi(EARGF(usage()));
			break;
		case 'a':
			nalt = atoi(EARGF(usage()));
			if(nalt > 10 || nalt < 0)
				nalt = 5;
			break;
		case 'p':
			flags |= PFLAG;
			break;
		case 'n':
			b = EARGF(usage());
			test = strchr(b, '?');
			if(test == nil)
				usage();
			lon = atof(b);
			lat = atof(test + 1);
			flags |= NFLAG;
			break;
		case 'w':
			flags |= WFLAG;
			break;
		default:
			usage();
	} ARGEND;

	if(initdraw(nil, nil, "gmap") < 0)
		sysfatal("initdraw failed: %r");

	pos = reallocg(nil, sizeof(Pos));
	pos->cur.lat = lat;
	pos->cur.lon = lon;
	pos->ocur = pos->cur;
	pos->dir = -1;
	pos->alt = alts[nalt];
	pos->str = nil;

	view = reallocg(nil, sizeof(Pos));
	view->cur.lat = lat;
	view->cur.lon = lon;
	view->ocur = view->cur;
	view->dir = -1;
	view->alt = alts[nalt];
	view->str = nil;

	backc = allocimagemix(display, DPaleyellow, DWhite);
	if(backc == nil)
		sysfatal("allocimagemix: %r");
	pbord = allocimagemix(display, DPaleyellow, DBlue);
	if(pbord == nil)
		sysfatal("allocimagemix: %r");
	pinne = allocimagemix(display, DBlue, DBlue);
	if(pinne == nil)
		sysfatal("allocimagemix: %r");
	textc = allocimagemix(display, DBlue, DBlue);
	if(textc == nil)
		sysfatal("allocimagemix: %r");
	waypc = allocimagemix(display, DBlue, DBlue);
	if(waypc == nil)
		sysfatal("allocimagemix: %r");

	eresized(0);

	einit(Emouse);

	Etimer = etimer(0, sec * 1000);

	while(1)
	{
		switch(eread(Emouse|Etimer, &e))
		{
			case Emouse:
				if(e.mouse.buttons & 1)
				{
					if(flags & NFLAG)
					{
						pos->cur = get_geo(view, bg, e.mouse.xy.x - screen->r.min.x -
									Dx(bg->bg->r) / 2 + bg->bg->r.min.x,
									-(e.mouse.xy.y - screen->r.min.y - Dy(bg->bg->r) / 2 +
									bg->bg->r.min.y));
						draw_nav(0);
					}
					break;
				}
				if(e.mouse.buttons & 4)
				{
					switch(epiemenuhit(3, &e.mouse, &menu))
					{
						case ZOOMIN:
							nalt--;
							if(nalt < 0)
								nalt = 10;
							draw_nav(0);
							break;
						case ZOOMOUT:
							nalt++;
							if(nalt > 10)
								nalt = 0;
							draw_nav(0);
							break;
						case REFRESH:
							draw_nav(0);
							break;
						case POS:
							flags ^= PFLAG;
							break;
						case WAYPOINT:
							flags ^= WFLAG;
							break;
						case CLR_WAYPOINT:
							if(posl != nil)
							{
								free_posl(posl);
								posl = nil;
							}
							break;
							draw_nav(0);
						case EXIT:
							goto free_all;
						default:
							break;
					}
				}
				break;
			default:
				draw_nav(0);
				break;
		}
		sleep(1);
	}

free_all:
	if(bg != nil)
		free_map(bg);
	if(pos != nil)
		free_pos(pos);
	if(view != nil)
		free_pos(view);
	if(posl != nil)
		free_posl(posl);
	freeimage(backc);
	freeimage(pbord);
	freeimage(pinne);
	freeimage(textc);
	freeimage(waypc);
	freeimage(backbuf);

	exits(0);
}
